# -*- coding: utf-8 -*-

import re
import unittest
import mock
import os
import tempfile
from contextlib import contextmanager
import time
import json
import logging


JIMBO = {
    'username': 'jimbo',
    'gpg_keyid': [],
    'approved_memberships': [],
    'cla_done': True,
    'groups': [],
    'timezone': 'Europe/Prague',
    'fullname': 'Jimbo',
    'email': 'jimbo@example.com',
    'login_time': time.time()
}


@contextmanager
def login(c, user):
    with c.session_transaction() as sess:
        sess['FLASK_FAS_OPENID_USER'] = user
    yield


def get_csrf_token(data):
    for line in data.split('\n'):
        if 'csrf_token' in line:
            m = re.search(r'value="([^"]+)', line)
            return m.groups()[0]
    return None


def create_project(owner):
    import poormanci
    p = poormanci.models.Project('pungi', 'Integration', owner,
                                 'pungi-on-pagure', 'http://pagure.io/', 'pagure_token',
                                 'pungi-on-jenkins', 'http://jenkins.io/', 'jenkins_token',
                                 'secret')
    poormanci.db_session.add(p)
    poormanci.db_session.commit()


class AppTestCase(unittest.TestCase):
    def setUp(self):
        self.db_fd, self.db_file = tempfile.mkstemp()
        conf = {
            'DATABASE': 'sqlite:///' + self.db_file,
            'TESTING': True,
        }
        import poormanci
        poormanci.app.config.update(conf)
        poormanci.connect_db()
        self.app = poormanci.app.test_client()
        poormanci.models.init_db(poormanci.app.config['DATABASE'])

    def tearDown(self):
        os.close(self.db_fd)
        os.unlink(self.db_file)

    def test_index_not_authenticated(self):
        rv = self.app.get('/')
        self.assertIn('You will have to log in', rv.data)
        self.assertEqual(200, rv.status_code)

    def test_index_authenticated(self):
        with mock.patch('flask.session', {'FLASK_FAS_OPENID_USER': JIMBO}):
            rv = self.app.get('/')
            self.assertIn('Jimbo', rv.data)
            self.assertIn('Log out', rv.data)
            self.assertIn('Nothing here yet', rv.data)
            self.assertEqual(200, rv.status_code)

    def test_can_add_project(self):
        with login(self.app, JIMBO):
            rv = self.app.get('/new')
            self.assertEqual(200, rv.status_code)

            data = {
                'csrf_token': get_csrf_token(rv.data),
                'name': 'pungi',
                'display_name': 'My Integration',
                'pagure_name': 'pungi-on-pagure',
                'jenkins_name': 'pungi-on-jenkins',
                'pagure_url': 'http://pagure.io',
                'jenkins_url': 'http://jenkins.io/',
                'pagure_token': 'pagure_token',
                'jenkins_token': 'jenkins_token',
            }
            rv = self.app.post('/new', data=data, follow_redirects=True)
            self.assertEqual(200, rv.status_code)
            self.assertIn('pungi', rv.data)
            self.assertIn('My Integration', rv.data)
            self.assertIn('http://pagure.io/pungi-on-pagure', rv.data)
            self.assertIn('http://jenkins.io/job/pungi-on-jenkins', rv.data)

    def test_can_not_add_duplicate_project(self):
        create_project('jimbo')
        with login(self.app, JIMBO):
            rv = self.app.get('/new')
            self.assertEqual(200, rv.status_code)

            data = {
                'csrf_token': get_csrf_token(rv.data),
                'name': 'pungi',
                'display_name': 'Yet another name',
                'pagure_name': 'pungi-on-pagure',
                'jenkins_name': 'pungi-on-jenkins',
                'pagure_url': 'http://pagure.io',
                'jenkins_url': 'http://jenkins.io/',
                'pagure_token': 'pagure_token',
                'jenkins_token': 'jenkins_token',
            }
            rv = self.app.post('/new', data=data, follow_redirects=True)
            self.assertEqual(200, rv.status_code)
            self.assertIn('This name is already used.', rv.data)

    def test_can_not_add_when_not_logged_in(self):
        rv = self.app.get('/new')
        self.assertEqual(302, rv.status_code)

    def test_can_edit_project(self):
        create_project('jimbo')

        with login(self.app, JIMBO):
            rv = self.app.get('/edit/pungi')
            self.assertEqual(200, rv.status_code)
            self.assertIn('Update pungi', rv.data)

            data = {
                'csrf_token': get_csrf_token(rv.data),
                'name': 'XXX',
                'display_name': 'YYY',
                'pagure_name': 'pungi-on-pagure',
                'jenkins_name': 'pungi-on-jenkins',
                'pagure_url': 'http://pagure.org',
                'jenkins_url': 'http://jenkins.org/',
                'pagure_token': 'pagure_token',
                'jenkins_token': 'jenkins_token',
            }
            rv = self.app.post('/edit/pungi', data=data, follow_redirects=True)
            self.assertEqual(200, rv.status_code)
            self.assertNotIn('XXX', rv.data)
            self.assertIn('YYY', rv.data)
            self.assertNotIn('Integration', rv.data)
            self.assertIn('http://pagure.org/pungi-on-pagure', rv.data)
            self.assertIn('http://jenkins.org/job/pungi-on-jenkins', rv.data)

    def test_can_not_edit_when_not_logged_in(self):
        rv = self.app.get('/edit/pungi')
        self.assertEqual(302, rv.status_code)

    def test_can_not_edit_when_not_owner(self):
        create_project('kevin')

        with login(self.app, JIMBO):
            rv = self.app.get('/edit/pungi')
            self.assertEqual(403, rv.status_code)

    def test_can_not_edit_when_not_existing(self):
        with login(self.app, JIMBO):
            rv = self.app.get('/edit/blah')
            self.assertEqual(404, rv.status_code)

    def test_can_delete_project(self):
        create_project('jimbo')
        with login(self.app, JIMBO):
            rv = self.app.get('/delete/pungi')
            data = {
                'csrf_token': get_csrf_token(rv.data),
            }
            rv = self.app.post('/delete/pungi', data=data, follow_redirects=True)
            self.assertEqual(200, rv.status_code)
            self.assertIn('Nothing here yet', rv.data)

    def test_can_not_delete_when_not_logged_in(self):
        rv = self.app.get('/delete/pungi')
        self.assertEqual(302, rv.status_code)

    def test_can_not_delete_when_not_owner(self):
        create_project('kevin')

        with login(self.app, JIMBO):
            rv = self.app.get('/delete/pungi')
            self.assertEqual(403, rv.status_code)

    def test_can_not_delete_not_existing(self):
        with login(self.app, JIMBO):
            rv = self.app.get('/delete/pungi')
            self.assertEqual(404, rv.status_code)

    @mock.patch('poormanci.lib.process_build')
    def test_jenkins_hook_ok(self, process_build):
        create_project('jimbo')
        rv = self.app.post('/hooks/secret/build-finished',
                           data=json.dumps({'name': 'pungi-on-jenkins', 'build': {'number': 15}}))
        self.assertEqual(204, rv.status_code)
        self.assertEqual(1, len(process_build.call_args_list))
        args, kwargs = process_build.call_args_list[0]
        self.assertEqual(4, len(args))
        import poormanci
        self.assertIs(args[0], poormanci.db_session)
        self.assertTrue(isinstance(args[1], logging.Logger))
        self.assertTrue(isinstance(args[2], poormanci.models.Project))
        self.assertEqual(15, args[3])

    @mock.patch('poormanci.lib.process_build')
    def test_jenkins_hook_bad_token(self, process_build):
        create_project('jimbo')
        rv = self.app.post('/hooks/foobar/build-finished',
                           data=json.dumps({'name': 'pungi-on-jenkins', 'build': {'number': 15}}))
        self.assertEqual(400, rv.status_code)
        self.assertEqual([], process_build.call_args_list)

    @mock.patch('poormanci.lib.process_build')
    def test_jenkins_hook_bad_data(self, process_build):
        create_project('jimbo')
        rv = self.app.post('/hooks/foobar/build-finished',
                           data=json.dumps({'name': 'pungi-on-jenkins', 'build': None}))
        self.assertEqual(400, rv.status_code)
        self.assertEqual([], process_build.call_args_list)

    @mock.patch('poormanci.lib.process_build')
    def test_jenkins_hook_unconfigured_project(self, process_build):
        create_project('jimbo')
        rv = self.app.post('/hooks/foobar/build-finished',
                           data=json.dumps({'name': 'asgard', 'build': {'number': 1}}))
        self.assertEqual(400, rv.status_code)
        self.assertEqual([], process_build.call_args_list)
